#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import aioboto3
import logging

import octobot_tentacles_manager.uploaders.uploader as uploader


class S3Uploader(uploader.Uploader):
    ENV_S3_API_KEY = "S3_API_KEY"
    ENV_S3_API_SECRET_KEY = "S3_API_SECRET_KEY"
    ENV_S3_ENDPOINT_URL = "S3_ENDPOINT_URL"
    ENV_S3_BUCKET_NAME = "S3_BUCKET_NAME"
    ENV_S3_REGION_NAME = "S3_REGION_NAME"

    S3_BOTO_KEY = "s3"

    def __init__(self):
        super().__init__()
        self._fix_loggers_level()
        self.s3_session: aioboto3.Session = aioboto3.Session()
        self.s3_api_key: str = os.getenv(S3Uploader.ENV_S3_API_KEY, None)
        self.s3_api_secret_key: str = os.getenv(S3Uploader.ENV_S3_API_SECRET_KEY, None)
        self.s3_endpoint_url: str = os.getenv(S3Uploader.ENV_S3_ENDPOINT_URL, None)
        self.s3_bucket_name: str = os.getenv(S3Uploader.ENV_S3_BUCKET_NAME, None)
        self.s3_region_name: str = os.getenv(S3Uploader.ENV_S3_REGION_NAME, None)
        if None in (self.s3_api_key, self.s3_api_secret_key, self.s3_endpoint_url,
                    self.s3_bucket_name, self.s3_region_name):
            raise TypeError("Some s3 environment variables are missing, please ensure that "
                            "S3_API_KEY, S3_API_SECRET_KEY, S3_BUCKET_NAME, S3_REGION_NAME "
                            "and S3_ENDPOINT_URL are defined.")

    @staticmethod
    def _fix_loggers_level():
        """
        Set aioboto3 and its components loggers level to Warning
        """
        logging.getLogger('aioboto3').setLevel(logging.WARNING)
        logging.getLogger('boto3').setLevel(logging.WARNING)
        logging.getLogger('botocore').setLevel(logging.WARNING)
        logging.getLogger('nose').setLevel(logging.WARNING)

    async def upload_file(self, upload_path: str, file_path: str, destination_file_name: str = None) -> int:
        """
        Upload file in s3
        :param upload_path: the upload path
        :param file_path: the file local path
        :param destination_file_name: the file name in s3 bucket (optional : default file_path basename)
        :return: the result of _upload
        """
        dest_file_name: str = destination_file_name \
            if destination_file_name is not None else os.path.basename(file_path)
        upload_file_url: str = f"{upload_path}{dest_file_name}"
        self.logger.info(f"Uploading {file_path} to s3 at {upload_file_url}...")
        return await self._upload(object_name=upload_file_url, local_file_path=file_path)

    async def upload_folder(self, upload_path: str, folder_path: str, destination_folder_name: str = None) -> int:
        """
        Upload recursive folder content in s3
        :param upload_path: the upload path
        :param folder_path: the folder local path
        :param destination_folder_name: the folder name in s3 bucket (optional : default folder_path basename)
        :return: the sum of all of _upload returns
        """
        dest_folder_name: str = destination_folder_name \
            if destination_folder_name is not None else os.path.basename(folder_path)
        upload_folder_url: str = f"{upload_path}{dest_folder_name}"
        return await self._upload_folder_content(upload_folder_url, folder_path)

    async def _upload_folder_content(self, upload_folder_url: str, folder_path: str) -> int:
        """
        Upload folder content in s3
        :param upload_folder_url: the folder upload path
        :param folder_path: the current folder path
        :return: the sum of all of _upload returns
        """
        error_count: int = 0
        for file_path in os.listdir(folder_path):
            if os.path.isdir(file_path):
                error_count += await self._upload_folder_content(upload_folder_url, folder_path)
            else:
                upload_file_url = f"{upload_folder_url}/{file_path}"
                self.logger.debug(f"Uploading {file_path} to s3 at {upload_file_url}...")
                error_count += await self._upload(object_name=upload_file_url,
                                                  local_file_path=os.path.join(folder_path, file_path))
        return error_count

    async def _upload(self, object_name: str, local_file_path: str) -> int:
        """
        Upload a file in the s3 bucket
        :param object_name: the object path
        :param local_file_path: the local file path
        :return: 0 if upload succeed else 1
        """
        async with self.s3_session.client(service_name=self.S3_BOTO_KEY,
                                          region_name=self.s3_region_name,
                                          endpoint_url=self.s3_endpoint_url,
                                          aws_access_key_id=self.s3_api_key,
                                          aws_secret_access_key=self.s3_api_secret_key) as s3:
            try:
                with open(local_file_path, 'rb') as file_content:
                    await s3.upload_fileobj(file_content, self.s3_bucket_name, object_name)
            except Exception as exception:
                self.logger.error(f"Failed to upload file on s3 bucket : {exception}")
                return 1
        return 0



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import packaging.version as packaging_version

import octobot_commons.logging as logging
import octobot_commons.tentacles_management as tentacles_management

import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.loaders as loaders


def get_installed_tentacles_modules() -> set:
    return set(tentacle for tentacle in loaders.get_tentacle_classes().values())


def get_tentacle_group(klass) -> str:
    return loaders.get_tentacle(klass).tentacle_group


def get_tentacle_version(klass) -> str:
    return loaders.get_tentacle(klass).version


def get_tentacle_origin_package(klass) -> str:
    return loaders.get_tentacle(klass).origin_package


def get_tentacle_module_name(klass) -> str:
    return loaders.get_tentacle(klass).name


def get_tentacle_classes_requirements(klass) -> list:
    requirements = loaders.get_tentacle(klass).extract_tentacle_requirements()
    classes = []
    for requirement in requirements:
        for tentacle_name in loaders.get_tentacles_classes_names_from_tentacle_module(requirement[0]):
            try:
                classes += [get_tentacle_class_from_string(tentacle_name)]
            except RuntimeError:
                # some tentacles can't be found this way (ex: util tentacles), ignore them
                pass
    return classes


def get_tentacle_resources_path(klass) -> str:
    return loaders.get_resources_path(klass)


def get_tentacle_documentation_path(klass) -> str:
    return loaders.get_documentation_file_path(klass)


def check_tentacle_version(version, name, origin_package, verbose=True) -> bool:
    logger = logging.get_logger("TentacleChecker")
    try:
        if origin_package == constants.DEFAULT_TENTACLES_PACKAGE:
            if packaging_version.Version(version) < packaging_version.Version(
                    constants.TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION) and verbose:
                logger.error(f"Incompatible tentacle {name}: version {version}, "
                             f"minimum expected: {constants.TENTACLE_CURRENT_MINIMUM_DEFAULT_TENTACLES_VERSION} "
                             f"this tentacle may not work properly. "
                             f"Please update your tentacles ('start.py -p update {name}' "
                             f"or 'start.py -p update all')")
                return False
    except Exception as e:
        if verbose:
            logger.error(f"Error when reading tentacle metadata: {e}")
    return True


def _load_tentacle_class(tentacle_name):
    # Lazy import of tentacles to let tentacles manager handle imports
    try:
        import octobot_evaluators.evaluators as evaluators
        import tentacles.Evaluator as tentacles_Evaluator
        if tentacle_class := tentacles_management.get_class_from_string(
            tentacle_name, evaluators.StrategyEvaluator,
            tentacles_Evaluator.Strategies, tentacles_management.evaluator_parent_inspection):
            return tentacle_class
        if tentacle_class := tentacles_management.get_class_from_string(
            tentacle_name, evaluators.TAEvaluator,
            tentacles_Evaluator.TA, tentacles_management.evaluator_parent_inspection):
            return tentacle_class
        if tentacle_class := tentacles_management.get_class_from_string(
            tentacle_name, evaluators.SocialEvaluator,
            tentacles_Evaluator.Social, tentacles_management.evaluator_parent_inspection):
            return tentacle_class
        if tentacle_class := tentacles_management.get_class_from_string(
            tentacle_name, evaluators.RealTimeEvaluator,
            tentacles_Evaluator.RealTime, tentacles_management.evaluator_parent_inspection):
            return tentacle_class
        if tentacle_class := tentacles_management.get_class_from_string(
            tentacle_name, evaluators.ScriptedEvaluator,
            tentacles_Evaluator.Scripted, tentacles_management.evaluator_parent_inspection):
            return tentacle_class
        import octobot_trading.modes as trading_modes
        import tentacles.Trading as tentacles_trading
        if tentacle_class := tentacles_management.get_class_from_string(
            tentacle_name, trading_modes.AbstractTradingMode,
            tentacles_trading.Mode, tentacles_management.trading_mode_parent_inspection):
            return tentacle_class
        import octobot_trading.exchanges as trading_exchanges
        if tentacle_class := tentacles_management.get_class_from_string(
            tentacle_name, trading_exchanges.AbstractExchange,
            tentacles_trading.Exchange, tentacles_management.default_parents_inspection):
            return tentacle_class
        try:
            import tentacles.Services.Interfaces.web_interface.plugins as web_plugins
            import tentacles.Services as tentacles_services
            if tentacle_class := tentacles_management.get_class_from_string(
                tentacle_name, web_plugins.AbstractWebInterfacePlugin,
                tentacles_services.Interfaces, tentacles_management.default_parents_inspection):
                return tentacle_class
        except ImportError:
            pass
        raise RuntimeError(f"Can't find tentacle: {tentacle_name}")
    except ImportError as e:
        raise ImportError(f"Can't import {e} module which is required to get associated "
                          f"tentacles classes") from e


def get_tentacle_class_from_string(tentacle_name, allow_cache=True):
    if allow_cache:
        try:
            return loaders.get_tentacle_class_from_name(tentacle_name)
        except KeyError:
            tentacle_class = _load_tentacle_class(tentacle_name)
            loaders.set_tentacle_class_by_name(tentacle_name, tentacle_class)
            return tentacle_class
    return _load_tentacle_class(tentacle_name)



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import octobot_tentacles_manager.models.artifact as artifact
import octobot_tentacles_manager.constants as constants


class TentaclePackage(artifact.Artifact):
    ARTIFACT_NAME = "tentacle package"
    ARTIFACT_SUFFIX = "package"

    def __init__(self, name=None):
        super().__init__(name)
        self.artifacts: list = []
        self.parse_path_in_name_if_any()

    def add_artifact(self, new_artifact: artifact.Artifact) -> None:
        """
        Add a new artifact to the package
        Used to manage the artifact name :
        - If only one artifact is present in the package -> use its name and version
        - If more artifacts are present, use the first artifact's origin package as name
        :param new_artifact: the artifact to add
        :return: None
        """
        self.artifacts.append(new_artifact)
        if len(self.artifacts) == 1:
            self.name = self.artifacts[0].get_name_with_version()
        elif len(self.artifacts) > 1:
            self.name = self.get_package_default_name()

    def parse_path_in_name_if_any(self) -> None:
        """
        Parse path in package name if any
        :return: None
        """
        if self.name \
                and len(os.path.split(self.name)) > 1 \
                and (split_path := os.path.split(self.name))[0]:
            self.output_path = split_path[0]
            self.name = split_path[1]

    def get_package_default_name(self) -> str:
        """
        :return: the default package name as 'origin_package@version_package'
        """
        return f"{self.artifacts[0].origin_package}{constants.ARTIFACT_VERSION_SEPARATOR}" \
               f"{self.version}_{self.ARTIFACT_SUFFIX}"

    def __str__(self):
        str_rep = f"{self.name} {TentaclePackage.ARTIFACT_NAME} ["
        if self.is_valid():
            return f"version: {self.version}]"
        else:
            return f"{str_rep}]"



#  Drakkar-Software OctoBot-Tentacles-Manager
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import os.path as path

import octobot_commons.logging as logging
import octobot_tentacles_manager.constants as constants
import octobot_tentacles_manager.models as models


def load_tentacle_with_metadata(tentacle_path: str):
    loaded_tentacles = _parse_all_tentacles(tentacle_path)
    _load_all_metadata(loaded_tentacles)
    return loaded_tentacles


def get_tentacles_from_package(tentacles, package_name: str):
    return [
        tentacle
        for tentacle in tentacles
        if tentacle.origin_package == package_name
    ]


def _load_all_metadata(tentacles):
    for tentacle in tentacles:
        try:
            tentacle.sync_initialize()
        except Exception as e:
            logging.get_logger(__name__).error(f"Error when loading {tentacle} ({tentacle.tentacle_path}) metadata: {e}")
            raise e


def _parse_all_tentacles(root: str):
    factory = models.TentacleFactory(root)
    return [
        factory.create_tentacle_from_type(tentacle_entry.name, tentacle_type)
        for tentacle_type in _get_tentacle_types(root)
        for tentacle_entry in os.scandir(path.join(root, tentacle_type.to_path()))
        if not (tentacle_entry.name == constants.PYTHON_INIT_FILE or
                tentacle_entry.name in constants.FOLDERS_BLACK_LIST)
    ]


def _get_tentacle_types(ref_tentacles_root):
    tentacle_types = []
    if path.isdir(ref_tentacles_root):
        _find_tentacles_type_in_directories(ref_tentacles_root, tentacle_types)
    return tentacle_types


def _find_tentacles_type_in_directories(ref_tentacles_root: os.DirEntry, tentacle_types: list, current_level=1):
    if current_level <= constants.TENTACLE_MAX_SUB_FOLDERS_LEVEL:
        for tentacle_type_entry in os.scandir(ref_tentacles_root):
            if tentacle_type_entry.is_dir():
                _explore_tentacle_dir(tentacle_type_entry, tentacle_types, current_level)


def _explore_tentacle_dir(tentacle_type_entry: os.DirEntry, tentacle_types: list, nesting_level: int):
    # full_tentacle_type_path is the path to the current directory starting from the most global Tentacle type without
    # anything before it:
    # ex: with a at tentacle_dir="downloaded_tentacles/xyz/Backtesting/importers" then
    # full_tentacle_type_path will be "Backtesting/importers"
    full_tentacle_type_path = tentacle_type_entry.path.split(path.sep)[-nesting_level:]
    if not _add_tentacle_type_if_is_valid(tentacle_type_entry, path.join(*full_tentacle_type_path), tentacle_types):
        # no tentacle in this folder, look into nested folders
        _find_tentacles_type_in_directories(tentacle_type_entry, tentacle_types, nesting_level + 1)


def _add_tentacle_type_if_is_valid(tentacle_type_entry: os.DirEntry,
                                   full_tentacle_type_path: str,
                                   tentacle_types: list):
    if _has_tentacle_in_direct_sub_directories(tentacle_type_entry):
        tentacle_types.append(models.TentacleType(full_tentacle_type_path))
        return True
    return False


def _has_tentacle_in_direct_sub_directories(directory_entry: os.DirEntry):
    return any(
        file_entry.name == constants.TENTACLE_METADATA
        for sub_directory_entry in os.scandir(directory_entry)
        if sub_directory_entry.is_dir()
        for file_entry in os.scandir(sub_directory_entry)
    )
